@@ -25,11 +25,12 @@ from linkedin_messaging.api_objects import (
     MessageAttachment,
     MessageCreate,
     ReactionSummary,
+    RealTimeEventStreamEvent,
     ThirdPartyMedia,
 )
 from mautrix.appservice import IntentAPI
 from mautrix.bridge import async_getter_lock, BasePortal, NotificationDisabler
-from mautrix.errors import MatrixError
+from mautrix.errors import MatrixError, MForbidden
 from mautrix.types import (
     AudioInfo,
     ContentURI,
@@ -77,6 +78,15 @@ try:
 except ImportError:
     decrypt_attachment = encrypt_attachment = None  # type: ignore
 
+
+class FakeLock:
+    async def __aenter__(self):
+        pass
+
+    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any):
+        pass
+
+
 StateBridge = EventType.find("m.bridge", EventType.Class.STATE)
 StateHalfShotBridge = EventType.find("uk.half-shot.bridge", EventType.Class.STATE)
 MediaInfo = Union[FileInfo, VideoInfo, AudioInfo, ImageInfo]
@@ -92,6 +102,7 @@ class Portal(DBPortal, BasePortal):
     backfill_lock: SimpleLock
     _dedup: Deque[URN]
     _send_locks: Dict[URN, asyncio.Lock]
+    _noop_lock: FakeLock = FakeLock()
 
     def __init__(
         self,
@@ -163,6 +174,13 @@ class Portal(DBPortal, BasePortal):
             self._send_locks[li_member_urn] = lock
         return lock
 
+    def optional_send_lock(self, li_member_urn: URN) -> Union[asyncio.Lock, FakeLock]:
+        try:
+            return self._send_locks[li_member_urn]
+        except KeyError:
+            pass
+        return self._noop_lock
+
     # endregion
 
     # region Properties
@@ -646,6 +664,13 @@ class Portal(DBPortal, BasePortal):
 
     # region Matrix event handling
 
+    async def _send_delivery_receipt(self, event_id: EventID) -> None:
+        if event_id and self.config["bridge.delivery_receipts"] and self.mxid:
+            try:
+                await self.az.intent.mark_read(self.mxid, event_id)
+            except Exception:
+                self.log.exception(f"Failed to send delivery receipt for {event_id}")
+
     async def handle_matrix_message(
         self,
         sender: "u.User",
@@ -715,6 +740,7 @@ class Portal(DBPortal, BasePortal):
                 index=0,
                 timestamp=datetime.now(),
             )
+            await self._send_delivery_receipt(event_id)
             self._dedup.append(resp.value.event_urn)
             await message.insert()
             return message
@@ -767,6 +793,66 @@ class Portal(DBPortal, BasePortal):
             MessageCreate(AttributedBody(), attachments=[attachment]),
         )
 
+    async def handle_matrix_redaction(
+        self,
+        sender: "u.User",
+        event_id: EventID,
+        redaction_event_id: EventID,
+    ):
+        if not self.mxid:
+            return
+
+        message = await DBMessage.get_by_mxid(event_id, self.mxid)
+        if message:
+            try:
+                await message.delete()
+                await sender.client.delete_message(
+                    self.li_thread_urn, message.li_message_urn
+                )
+            except Exception:
+                self.log.exception("Delete message failed")
+
+        reaction = await DBReaction.get_by_mxid(event_id, self.mxid)
+        if reaction:
+            try:
+                await reaction.delete()
+                await sender.client.remove_emoji_reaction(
+                    self.li_thread_urn, reaction.li_message_urn, emoji=reaction.reaction
+                )
+            except Exception:
+                self.log.exception("Removing reaction failed")
+
+        await self._send_delivery_receipt(redaction_event_id)
+
+    async def handle_matrix_reaction(
+        self,
+        sender: "u.User",
+        event_id: EventID,
+        reacting_to: EventID,
+        reaction: str,
+    ):
+        assert sender.li_member_urn
+        assert self.mxid
+        async with self.require_send_lock(sender.li_member_urn):
+            message = await DBMessage.get_by_mxid(reacting_to, self.mxid)
+            if not message:
+                self.log.debug(f"Ignoring reaction to unknown event {reacting_to}")
+                return
+
+            await sender.client.add_emoji_reaction(
+                self.li_thread_urn, message.li_message_urn, reaction
+            )
+            await DBReaction(
+                mxid=event_id,
+                mx_room=message.mx_room,
+                li_message_urn=message.li_message_urn,
+                li_receiver_urn=self.li_receiver_urn,
+                li_sender_urn=sender.li_member_urn,
+                reaction=reaction,
+            ).insert()
+
+        await self._send_delivery_receipt(event_id)
+
     # endregion
 
     # region LinkedIn event handling
@@ -804,7 +890,10 @@ class Portal(DBPortal, BasePortal):
         message: ConversationEvent,
     ):
         try:
-            await self._handle_linkedin_message(source, sender, message)
+            if message.event_content.message_event.recalled_at:
+                await self._handle_linkedin_message_deletion(sender, message)
+            else:
+                await self._handle_linkedin_message(source, sender, message)
         except Exception as e:
             self.log.exception(
                 f"Error handling LinkedIn message {message.entity_urn}: {e}"
@@ -821,122 +910,127 @@ class Portal(DBPortal, BasePortal):
         li_message_urn = message.entity_urn
 
         # Check in-memory queue for duplicates
+        message_exists = False
+        event_ids: List[EventID] = []
         async with self.require_send_lock(sender.li_member_urn):
             if li_message_urn in self._dedup:
                 self.log.trace(
                     f"Not handling message {li_message_urn}, found ID in dedup queue"
                 )
+                # Return here, because it is in the process of being handled.
                 return
             self._dedup.appendleft(li_message_urn)
 
             # Check database for duplicates
-            dbm = await DBMessage.get_by_li_message_urn(
+            dbm = await DBMessage.get_all_by_li_message_urn(
                 li_message_urn, self.li_receiver_urn
             )
-            if dbm:
+            if len(dbm) > 0:
                 self.log.debug(
                     f"Not handling message {li_message_urn}, found duplicate in "
                     "database."
                 )
-                return
-
-        self.log.trace("LinkedIn event content: %s", message)
-        if not self.mxid:
-            mxid = await self.create_matrix_room(source)
-            if not mxid:
-                # Failed to create
-                return
-        if not await self._bridge_own_message_pm(
-            source, sender, f"message {li_message_urn}"
-        ):
-            return
+                # Don't return here because we may need to update the reactions.
+                message_exists = True
+                event_ids = [dbm.mxid for dbm in sorted(dbm, key=lambda m: m.index)]
 
         intent = sender.intent_for(self)
+        if not message_exists:
+            self.log.trace("LinkedIn event content: %s", message)
+            if not self.mxid:
+                mxid = await self.create_matrix_room(source)
+                if not mxid:
+                    # Failed to create
+                    return
+            if not await self._bridge_own_message_pm(
+                source, sender, f"message {li_message_urn}"
+            ):
+                return
 
-        if (
-            self._backfill_leave is not None
-            and self.li_other_user_urn != sender.li_member_urn
-            and intent != sender.intent
-            and intent not in self._backfill_leave
-        ):
-            self.log.debug(
-                "Adding %s's default puppet to room for backfilling", sender.mxid
-            )
-            await self.main_intent.invite_user(self.mxid, intent.mxid)
-            await intent.ensure_joined(self.mxid)
-            self._backfill_leave.add(intent)
-
-        message_event = message.event_content.message_event
-        timestamp = message.created_at
-
-        event_ids = []
-
-        # Handle subject
-        if message_event.subject:
-            event_ids.append(
-                await self._send_message(
-                    intent,
-                    linkedin_subject_to_matrix(message_event.subject),
-                    timestamp=timestamp,
+            if (
+                self._backfill_leave is not None
+                and self.li_other_user_urn != sender.li_member_urn
+                and intent != sender.intent
+                and intent not in self._backfill_leave
+            ):
+                self.log.debug(
+                    "Adding %s's default puppet to room for backfilling", sender.mxid
                 )
-            )
+                await self.main_intent.invite_user(self.mxid, intent.mxid)
+                await intent.ensure_joined(self.mxid)
+                self._backfill_leave.add(intent)
 
-        # Handle attachments
-        event_ids.extend(
-            await self._handle_linkedin_attachments(
-                source, intent, timestamp, message_event.attachments
-            )
-        )
+            message_event = message.event_content.message_event
+            timestamp = message.created_at
 
-        # Handle custom content
-        if message_event.custom_content:
-            if message_event.custom_content.third_party_media:
-                event_ids.extend(
-                    await self._handle_linkedin_third_party_media(
-                        source,
-                        intent,
-                        timestamp,
-                        message_event.custom_content.third_party_media,
-                    )
-                )
+            event_ids = []
 
-            # Handle InMail message text
-            if message_event.custom_content.sp_inmail_content:
+            # Handle subject
+            if message_event.subject:
                 event_ids.append(
                     await self._send_message(
                         intent,
-                        await linkedin_spinmail_to_matrix(
-                            message_event.custom_content.sp_inmail_content
-                        ),
+                        linkedin_subject_to_matrix(message_event.subject),
                         timestamp=timestamp,
                     )
                 )
 
-        # Handle the normal message text itself
-        if message_event.attributed_body and message_event.attributed_body.text:
-            content = await linkedin_to_matrix(message_event.attributed_body)
-            event_ids.append(
-                await self._send_message(intent, content, timestamp=timestamp)
+            # Handle attachments
+            event_ids.extend(
+                await self._handle_linkedin_attachments(
+                    source, intent, timestamp, message_event.attachments
+                )
             )
-            # TODO (#55) error handling
 
-        event_ids = [event_id for event_id in event_ids if event_id]
-        if len(event_ids) == 0:
-            return
+            # Handle custom content
+            if message_event.custom_content:
+                if message_event.custom_content.third_party_media:
+                    event_ids.extend(
+                        await self._handle_linkedin_third_party_media(
+                            source,
+                            intent,
+                            timestamp,
+                            message_event.custom_content.third_party_media,
+                        )
+                    )
 
-        # Save all of the messages in the database.
-        self.log.debug(f"Handled LinkedIn message {li_message_urn} -> {event_ids}")
-        await DBMessage.bulk_create(
-            li_message_urn=li_message_urn,
-            li_thread_urn=self.li_thread_urn,
-            li_sender_urn=sender.li_member_urn,
-            li_receiver_urn=self.li_receiver_urn,
-            mx_room=self.mxid,
-            timestamp=timestamp,
-            event_ids=event_ids,
-        )
-        # TODO (#48)
-        # await self._send_delivery_receipt(event_ids[-1])
+                # Handle InMail message text
+                if message_event.custom_content.sp_inmail_content:
+                    event_ids.append(
+                        await self._send_message(
+                            intent,
+                            await linkedin_spinmail_to_matrix(
+                                message_event.custom_content.sp_inmail_content
+                            ),
+                            timestamp=timestamp,
+                        )
+                    )
+
+            # Handle the normal message text itself
+            if message_event.attributed_body and message_event.attributed_body.text:
+                content = await linkedin_to_matrix(message_event.attributed_body)
+                event_ids.append(
+                    await self._send_message(intent, content, timestamp=timestamp)
+                )
+                # TODO (#55) error handling
+
+            event_ids = [event_id for event_id in event_ids if event_id]
+            if len(event_ids) == 0:
+                return
+
+            # Save all of the messages in the database.
+            self.log.debug(f"Handled LinkedIn message {li_message_urn} -> {event_ids}")
+            await DBMessage.bulk_create(
+                li_message_urn=li_message_urn,
+                li_thread_urn=self.li_thread_urn,
+                li_sender_urn=sender.li_member_urn,
+                li_receiver_urn=self.li_receiver_urn,
+                mx_room=self.mxid,
+                timestamp=timestamp,
+                event_ids=event_ids,
+            )
+            await self._send_delivery_receipt(event_ids[-1])
+        # end if message_exists
 
         # Handle reactions
         reaction_summaries = sorted(
@@ -946,42 +1040,79 @@ class Portal(DBPortal, BasePortal):
         )
         reaction_event_id = event_ids[-1]  # react to the last event
         for reaction_summary in reaction_summaries:
-            await self.handle_reaction_summary(
-                intent,
+            await self._handle_reaction_summary(
                 li_message_urn,
-                sender,
+                source,
                 reaction_event_id,
                 reaction_summary,
             )
 
-    async def handle_reaction_summary(
+    async def _handle_linkedin_message_deletion(
         self,
-        intent: IntentAPI,
-        li_message_urn: URN,
         sender: "p.Puppet",
+        message: ConversationEvent,
+    ):
+        if not self.mxid or not self.li_receiver_urn:
+            return
+        for db_message in await DBMessage.get_all_by_li_message_urn(
+            message.entity_urn, self.li_receiver_urn
+        ):
+            try:
+                await sender.intent_for(self).redact(
+                    db_message.mx_room,
+                    db_message.mxid,
+                    timestamp=message.created_at,
+                )
+            except MForbidden:
+                await self.main_intent.redact(
+                    db_message.mx_room,
+                    db_message.mxid,
+                    timestamp=message.created_at,
+                )
+            await db_message.delete()
+
+    async def _handle_reaction_summary(
+        self,
+        li_message_urn: URN,
+        source: "u.User",
         reaction_event_id: EventID,
         reaction_summary: ReactionSummary,
-    ) -> Optional[EventID]:
-        if not reaction_summary.emoji:
-            return None
+    ) -> List[EventID]:
+        if not reaction_summary.emoji or not source.client:
+            return []
 
         assert self.mxid
         assert self.li_receiver_urn
 
-        # TODO (#32) figure out how many reactions should be added
-        mxid = await intent.react(self.mxid, reaction_event_id, reaction_summary.emoji)
-        self.log.debug(
-            f"Reacted to {reaction_event_id} with {reaction_summary.emoji}, got {mxid}"
-        )
-        await DBReaction(
-            mxid=mxid,
-            mx_room=self.mxid,
-            li_message_urn=li_message_urn,
-            li_receiver_urn=self.li_receiver_urn,
-            li_sender_urn=sender.li_member_urn,
-            reaction=reaction_summary.emoji,
-        ).insert()
-        return mxid
+        emoji = reaction_summary.emoji
+        reactors = await source.client.get_reactors(li_message_urn, emoji)
+
+        mxids = []
+        for reactor in reactors.elements:
+            sender = await p.Puppet.get_by_li_member_urn(reactor.reactor_urn)
+            intent = sender.intent_for(self)
+
+            # TODO (#33) figure out how many reactions should be added
+            mxid = await intent.react(
+                self.mxid, reaction_event_id, reaction_summary.emoji
+            )
+            mxids.append(mxid)
+
+            self.log.debug(
+                f"{sender.mxid} reacted to {reaction_event_id} with "
+                "{reaction_summary.emoji}, got {mxid}."
+            )
+
+            await DBReaction(
+                mxid=mxid,
+                mx_room=self.mxid,
+                li_message_urn=li_message_urn,
+                li_receiver_urn=self.li_receiver_urn,
+                li_sender_urn=sender.li_member_urn,
+                reaction=reaction_summary.emoji,
+            ).insert()
+
+        return mxids
 
     async def _handle_linkedin_attachments(
         self,
@@ -1112,4 +1243,89 @@ class Portal(DBPortal, BasePortal):
             decryption_info.url = url
         return url, info, decryption_info
 
+    async def handle_linkedin_reaction_add(
+        self,
+        source: "u.User",
+        sender: "p.Puppet",
+        event: RealTimeEventStreamEvent,
+    ):
+        assert event.event_urn
+        assert self.li_receiver_urn
+        reaction = event.reaction_summary.emoji
+        # Make up a URN for the reacton for dedup purposes
+        dedup_id = URN(
+            f"({event.event_urn.id_str()},{sender.li_member_urn.id_str()},{reaction})"
+        )
+        async with self.optional_send_lock(sender.li_member_urn):
+            if dedup_id in self._dedup:
+                return
+            self._dedup.appendleft(dedup_id)
+
+            # Check database for duplicates
+            dbr = await DBReaction.get_by_li_message_urn_and_emoji(
+                event.event_urn,
+                self.li_receiver_urn,
+                sender.li_member_urn,
+                reaction,
+            )
+            if dbr:
+                self.log.debug(
+                    f"Not handling reaction {reaction} to {event.event_urn}, found "
+                    "duplicate in database."
+                )
+                return
+
+        if not await self._bridge_own_message_pm(
+            source, sender, f"reaction to {event.event_urn}"
+        ):
+            return
+
+        intent = sender.intent_for(self)
+
+        message = await DBMessage.get_by_li_message_urn(
+            event.event_urn, self.li_receiver_urn
+        )
+        if not message:
+            self.log.debug(f"Ignoring reaction to unknown message {event.event_urn}")
+            return
+
+        mxid = await intent.react(message.mx_room, message.mxid, reaction)
+        self.log.debug(f"Reacted to {message.mxid}, got {mxid}")
+
+        await DBReaction(
+            mxid=mxid,
+            mx_room=message.mx_room,
+            li_message_urn=message.li_message_urn,
+            li_receiver_urn=self.li_receiver_urn,
+            li_sender_urn=sender.li_member_urn,
+            reaction=reaction,
+        ).insert()
+        self._dedup.remove(dedup_id)
+
+    async def handle_linkedin_reaction_remove(
+        self,
+        source: "u.User",
+        sender: "p.Puppet",
+        event: RealTimeEventStreamEvent,
+    ):
+        if (
+            not self.mxid
+            or not self.li_receiver_urn
+            or not event.event_urn
+            or not event.reaction_summary
+        ):
+            return
+        reaction = await DBReaction.get_by_li_message_urn_and_emoji(
+            event.event_urn,
+            self.li_receiver_urn,
+            sender.li_member_urn,
+            event.reaction_summary.emoji,
+        )
+        if reaction:
+            try:
+                await sender.intent_for(self).redact(reaction.mx_room, reaction.mxid)
+            except MForbidden:
+                await self.main_intent.redact(reaction.mx_room, reaction.mxid)
+            await reaction.delete()
+
     # endregion
